import os, sys

sys.path.append(os.path.dirname(__file__))

from copy import deepcopy
from typing import Any, Dict, Optional, Union

from tianshou.data import Batch, ReplayBuffer, to_numpy, to_torch_as
from tianshou.policy import DQNPolicy

import torch
from tianshou.utils.net.common import MLP
import torch.nn as nn
import numpy as np
from torch.nn.functional import softplus
from torch.distributions import Normal, Independent
import torch.nn.functional as F
import math
import random
from itertools import combinations
from other import make_path_if_not_exist
import matplotlib.pyplot as plt
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE
import seaborn as sns
from scipy.spatial.distance import cdist

from zqmtool.other import to_np, to_th


class GoalPosNegPairSampler():
    def __init__(self, goal_set, goal_state_set, len_step, len_time_window):
        self.goal_set = goal_set

        dist_matrix = cdist(goal_state_set, goal_state_set)
        self.pos_gid_pairs = np.stack(np.where(dist_matrix < len_step * len_time_window), axis=-1)
        self.neg_gid_pairs = np.stack(np.where(dist_matrix >= len_step * len_time_window), axis=-1)
        assert np.amax(
            np.linalg.norm(goal_state_set[self.pos_gid_pairs[:, 0]] - goal_state_set[self.pos_gid_pairs[:, 1]], axis=1)) \
               < len_step * len_time_window
        assert np.amin(
            np.linalg.norm(goal_state_set[self.neg_gid_pairs[:, 0]] - goal_state_set[self.neg_gid_pairs[:, 1]], axis=1)) \
               >= len_step * len_time_window

    def sample(self, batch_size):
        pos_gid_samples = rnd_sample_rows(self.pos_gid_pairs, n=batch_size).copy()
        neg_gid_samples = rnd_sample_rows(self.neg_gid_pairs, n=batch_size).copy()
        pos0 = self.goal_set[pos_gid_samples[:, 0]].copy()
        pos1 = self.goal_set[pos_gid_samples[:, 1]].copy()
        neg0 = self.goal_set[neg_gid_samples[:, 0]].copy()
        neg1 = self.goal_set[neg_gid_samples[:, 1]].copy()
        return pos0, pos1, neg0, neg1


def rnd_sample_rows(A, n):
    # A[np.random.choice(A.shape[0], 2, replace=False), :] # without replacement
    return A[np.random.randint(A.shape[0], size=n), :].copy()  # with replacement


def update_beta(method, beta, loss_infomin=None, acc=None, min_val=None, set_beta=None):
    def add_beta(beta):
        return min(beta + step_val, max_val)

    def reduce_beta(beta):
        return max(beta - step_val, min_val)

    max_val = 1.
    step_val = 1e-5

    if set_beta is not None:
        return min(beta + step_val, set_beta)

    if method.lower() == 'InfoGoal'.lower():
        min_val = 0.01
        if beta < min_val:
            return add_beta(beta)
        return add_beta(beta) if acc > 0.90 else reduce_beta(beta)
    elif method.lower() == 'InfoBot'.lower():
        return 0.01
    else:
        return beta


def reparameterization(params, z_dim, num_sample=1):
    mu, sigma = params[:, :z_dim], params[:, z_dim:]
    sigma = softplus(sigma) + 1e-3  # Make sigma always positive
    if num_sample == 1:
        z = Independent(Normal(loc=mu, scale=sigma), 1).rsample()
    else:
        z = Independent(Normal(loc=mu, scale=sigma), 1).rsample(sample_shape=(num_sample,))
    return z, mu, sigma


def infomax_loss_fn_two_goal_repre(z0, z1):
    return F.l1_loss(z0, z1, reduction='none')


def infomax_loss_fn(z_p0, z_p1, z_n0, z_n1, method='our', W=None):
    if method == 'our':
        z_pos = infomax_loss_fn_two_goal_repre(z_p0, z_p1)
        z_neg = infomax_loss_fn_two_goal_repre(z_n0, z_n1)

        z_pos = torch.mean(z_pos, dim=-1)
        z_neg = torch.mean(z_neg, dim=-1)
        loss = torch.exp(z_pos - z_neg)
        loss = torch.mean(loss)

        acc = torch.mean(torch.as_tensor(z_pos < z_neg, dtype=torch.float32))
        return loss, acc
    elif method == 'wmse':
        num_goal = int(len(z_p0))
        z = torch.cat([z_p0, z_p1], dim=0)
        z = W(z)
        z_a = z[:num_goal]
        z_p = z[num_goal:]
        loss = (z_a - z_p).pow(2).mean()
        return loss, torch.FloatTensor([0.])


def infomin_loss_fn(mu, std):
    loss = -0.5 * (1 + 2 * std.log() - mu.pow(2) - std.pow(2)).sum(1).mean().div(math.log(2))
    return loss


# def draw_scatter_emb(net, save_path, env_info, env, times, device, show_goal_id_set=None, enlarge_goal_id=None,
#                      show_legend=True, lw=4):
#     import matplotlib as mpl
#
#     make_path_if_not_exist(os.path.dirname(save_path))
#     mpl.rcParams['axes.linewidth'] = lw
#
#     SMALL_SIZE = 6 * times
#     MEDIUM_SIZE = 8 * times
#
#     plt.rc('pdf', fonttype=42)
#     plt.rc('ps', fonttype=42)
#     plt.rc('font', family='serif')
#     plt.rc('font', size=MEDIUM_SIZE)  # controls default text sizes
#     plt.rc('axes', titlesize=MEDIUM_SIZE)  # fontsize of the axes title
#     plt.rc('axes', labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
#     plt.rc('xtick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
#     plt.rc('ytick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
#     plt.rc('legend', fontsize=SMALL_SIZE)  # legend fontsize
#     plt.rc('figure', titlesize=MEDIUM_SIZE)  # fontsize of the figure title
#
#     goal_info_dict = env_info['goal_info_dict']
#     c_dict = env_info['c_dict']
#
#     cur_goal_ids = []
#     cur_goals = []
#     cur_colors = []
#     cur_modes = []
#     cur_sizes = []
#     for cur_goal_id in goal_info_dict.keys():
#         if show_goal_id_set is not None:
#             if cur_goal_id not in show_goal_id_set:
#                 continue
#         cur_mode = 'training' if cur_goal_id in env.training_goal_ids else 'test'
#         if (enlarge_goal_id is not None) and (cur_mode == 1) and (cur_goal_id != enlarge_goal_id): continue
#         cur_modes.append(cur_mode)
#         cur_goal_ids.append(cur_goal_id)
#         cur_colors.append(goal_info_dict[cur_goal_id]['room'])
#         if cur_goal_id == enlarge_goal_id:
#             cur_sizes.append(400)
#         else:
#             cur_sizes.append(200)
#
#     with torch.no_grad():
#         z, m, s = net.encode_fn(cur_goal_ids, type='goal', num_sample=100)
#         z_orig_shape = z.shape
#         if len(z_orig_shape) == 2:
#             z_orig_shape = (1, *z_orig_shape)
#         num_repeat = z_orig_shape[0]
#         num_goal = z_orig_shape[1]
#         num_dim = z_orig_shape[2]
#
#         cur_colors_for_z = cur_colors * num_repeat  # [cur_color for cur_color in cur_colors for i in range(num_repeat)]
#         cur_modes_for_z = cur_modes * num_repeat  # [cur_mode for cur_mode in cur_modes for i in range(num_repeat)]
#         cur_sizes_for_z = cur_sizes * num_repeat  # [cur_size for cur_size in cur_sizes for i in range(num_repeat)]
#
#         cur_colors_for_m = cur_colors
#         cur_modes_for_m = cur_modes
#         cur_sizes_for_m = cur_sizes
#
#         # cur_colors_for_z = np.asarray(cur_colors)[None, ...].repeat(z_orig_shape[0], axis=0).reshape(-1, 1)
#         # cur_modes_for_z = np.asarray(cur_modes)[None, ...].repeat(z_orig_shape[0], axis=0).reshape(-1, 1)
#         # cur_sizes_for_z = np.asarray(cur_sizes)[None, ...].repeat(z_orig_shape[0], axis=0).reshape(-1, 1)
#
#         # cur_colors_for_m = np.asarray(cur_colors).reshape(-1, 1)
#         # cur_modes_for_m = np.asarray(cur_modes).reshape(-1, 1)
#         # cur_sizes_for_m = np.asarray(cur_sizes).reshape(-1,1)
#
#         z = z.reshape(-1, num_dim).cpu().numpy()
#         z_2d = z if num_dim == 2 else PCA(n_components=2).fit_transform(z)
#         m_2d = np.mean(z_2d.reshape(z_orig_shape[0], z_orig_shape[1], -1), axis=0)
#
#         z_2d_w_color = [[z2i[0], z2i[1], cur_color_for_z, cur_mode_for_z, cur_size_for_z] for
#                         z2i, cur_color_for_z, cur_mode_for_z, cur_size_for_z in
#                         zip(z_2d, cur_colors_for_z, cur_modes_for_z, cur_sizes_for_z)]
#         m_2d_w_color = [[m2i[0], m2i[1], cur_color_for_m, cur_mode_for_m, cur_size_for_m] for
#                         m2i, cur_color_for_m, cur_mode_for_m, cur_size_for_m in
#                         zip(m_2d, cur_colors_for_m, cur_modes_for_m, cur_sizes_for_m)]
#
#         # z_2d_w_color = np.concatenate([z_2d, cur_colors_for_z, cur_modes_for_z, cur_sizes_for_z], axis=-1)
#         # m_2d_w_color = np.concatenate([m_2d, cur_colors_for_m, cur_modes_for_m, cur_sizes_for_m], axis=-1)
#
#     z_pd = pd.DataFrame(z_2d_w_color, columns=['x', 'y', 'room', 'mode', 'size'])
#     m_pd = pd.DataFrame(m_2d_w_color, columns=['x', 'y', 'room', 'mode', 'size'])
#
#     # m_pd = m_pd.replace({'mode': {0: 'training', 1: 'test'}})
#     # m_pd = m_pd.replace({'room': {0: 'A', 1: 'B', 2: 'C', 3: 'D'}})
#
#     ax = plt.gca()
#     sns.scatterplot(ax=ax, data=z_pd, x='x', y='y',
#                     alpha=0.1, style='mode', markers={'training': 'o', 'test': '*'}, s=800,
#                     hue='room', palette=c_dict, legend=False)
#     sns.scatterplot(ax=ax, data=m_pd, x='x', y='y',
#                     style='mode', markers={'training': 'o', 'test': '*'}, s=800,
#                     hue='room', palette=c_dict, legend=show_legend)
#
#     # for label, (marker, s) in {'training goal': ('o', 200), 'test goal': ('*', 400)}.items():
#     #     plt.scatter([], [], s=s, marker=marker, label=label, facecolors='none', edgecolors='k')
#     if show_legend:
#         plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), shadow=True, markerscale=2)  # ,
#
#     # ax.axis('off')
#     # ax.set_aspect('auto')
#     ax.set_xticks([])
#     ax.set_yticks([])
#     ax.set_xlabel('')
#     ax.set_ylabel('')
#     plt.savefig(save_path, bbox_inches='tight')
#     plt.close()
def draw_scatter_emb(z, save_path, times, goal_info, show_legend=True, lw=4, enlarge_goal_id=None):
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    from sklearn.decomposition import PCA
    import seaborn as sns
    import pandas as pd

    make_path_if_not_exist(os.path.dirname(save_path))
    mpl.rcParams['axes.linewidth'] = lw

    SMALL_SIZE = 6 * times
    MEDIUM_SIZE = 8 * times

    plt.rc('pdf', fonttype=42)
    plt.rc('ps', fonttype=42)
    plt.rc('font', family='serif')
    plt.rc('font', size=MEDIUM_SIZE)  # controls default text sizes
    plt.rc('axes', titlesize=MEDIUM_SIZE)  # fontsize of the axes title
    plt.rc('axes', labelsize=MEDIUM_SIZE)  # fontsize of the x and y labels
    plt.rc('xtick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
    plt.rc('ytick', labelsize=SMALL_SIZE)  # fontsize of the tick labels
    plt.rc('legend', fontsize=SMALL_SIZE)  # legend fontsize
    plt.rc('figure', titlesize=MEDIUM_SIZE)  # fontsize of the figure title

    cur_goal_ids = []
    cur_goals = []
    cur_colors = []
    cur_modes = []
    cur_sizes = []
    for cur_goal_id in goal_info.keys():
        cur_mode = goal_info[cur_goal_id]['mode']
        cur_modes.append(cur_mode)
        cur_goal_ids.append(cur_goal_id)
        cur_colors.append(goal_info[cur_goal_id]['room'])
        if cur_goal_id == enlarge_goal_id:
            cur_sizes.append(400)
        else:
            cur_sizes.append(200)

    with torch.no_grad():
        z_orig_shape = z.shape
        if len(z_orig_shape) == 2:
            z_orig_shape = (1, *z_orig_shape)
        num_repeat = z_orig_shape[0]
        num_goal = z_orig_shape[1]
        num_dim = z_orig_shape[2]

        cur_colors_for_z = cur_colors * num_repeat  # [cur_color for cur_color in cur_colors for i in range(num_repeat)]
        cur_modes_for_z = cur_modes * num_repeat  # [cur_mode for cur_mode in cur_modes for i in range(num_repeat)]
        cur_sizes_for_z = cur_sizes * num_repeat  # [cur_size for cur_size in cur_sizes for i in range(num_repeat)]

        cur_colors_for_m = cur_colors
        cur_modes_for_m = cur_modes
        cur_sizes_for_m = cur_sizes

        z = z.reshape(-1, num_dim).cpu().numpy()
        z_2d = z if num_dim == 2 else PCA(n_components=2).fit_transform(z)
        m_2d = np.mean(z_2d.reshape(z_orig_shape[0], z_orig_shape[1], -1), axis=0)

        z_2d_w_color = [[z2i[0], z2i[1], cur_color_for_z, cur_mode_for_z, cur_size_for_z] for
                        z2i, cur_color_for_z, cur_mode_for_z, cur_size_for_z in
                        zip(z_2d, cur_colors_for_z, cur_modes_for_z, cur_sizes_for_z)]
        m_2d_w_color = [[m2i[0], m2i[1], cur_color_for_m, cur_mode_for_m, cur_size_for_m] for
                        m2i, cur_color_for_m, cur_mode_for_m, cur_size_for_m in
                        zip(m_2d, cur_colors_for_m, cur_modes_for_m, cur_sizes_for_m)]

    z_pd = pd.DataFrame(z_2d_w_color, columns=['x', 'y', 'room', 'mode', 'size'])
    m_pd = pd.DataFrame(m_2d_w_color, columns=['x', 'y', 'room', 'mode', 'size'])

    ax = plt.gca()
    sns.scatterplot(ax=ax, data=z_pd, x='x', y='y',
                    alpha=0.1, style='mode', markers={'training': 'o', 'test': '*'}, s=800,
                    hue='room', palette="tab10", legend=False)
    sns.scatterplot(ax=ax, data=m_pd, x='x', y='y',
                    style='mode', markers={'training': 'o', 'test': '*'}, s=800,
                    hue='room', palette="tab10", legend=show_legend)

    if show_legend:
        plt.legend(loc='center left', bbox_to_anchor=(1, 0.5), shadow=True, markerscale=2)

    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_xlabel('')
    ax.set_ylabel('')
    plt.savefig(save_path, bbox_inches='tight')
    plt.close()


def save_learned_gr_for_im(save_path, algo_name, goal_info, enc, device, set_beta, set_win):
    with torch.no_grad():
        for g_im_key in goal_info[0]['representation']['im'].keys():
            g_im_w_rnd = to_np([goal_info[g_id]['representation']['im'][g_im_key] for g_id in goal_info.keys()])
            inp = to_th(g_im_w_rnd, device=device)
            params = enc(inp)
            z, m, s = reparameterization(params=params, z_dim=32)
            m = to_np(m.cpu())
            s = to_np(s.cpu())
            for g_id in goal_info.keys():
                goal_info[g_id]['representation'][algo_name][f'{g_im_key}_beta{set_beta}_win{set_win}'] = \
                    np.concatenate([m[g_id], s[g_id]])

            save_path_ = os.path.join(save_path, f'{g_im_key}.png')
            make_path_if_not_exist(os.path.dirname(save_path))
            z, m, s = reparameterization(params=params, z_dim=32, num_sample=100)
            draw_scatter_emb(
                z=z, goal_info=goal_info,
                save_path=save_path_,
                times=2, show_legend=False,
            )


class Base:

    def __init__(self, net, optim, batch_size, positive_instance_pairs):
        self.net = net
        self.optim = optim
        self.device = net.device
        self.batch_size = batch_size
        self.positive_instance_pairs = positive_instance_pairs
        self.len_positive_instance_pairs = None
        self.pos_pair_set = None
        self.neg_pair_set = None

    def divide_pos_neg_pairs(self, ):
        if (self.len_positive_instance_pairs is not None) and \
                (len(self.positive_instance_pairs) == self.len_positive_instance_pairs):
            return

        self.len_positive_instance_pairs = len(self.positive_instance_pairs)

        # get z_a
        a = []
        for pair in self.positive_instance_pairs:
            a += list(pair)
        a = set(a)

        pos_pair_set = []
        for ai in a:
            pos = []
            for pair in self.positive_instance_pairs:
                if ai in pair:
                    pos += list(pair)
            pos = set(pos)

            for posi in pos:
                pos_pair_set.append([ai, posi])

        neg_pair_set = []
        for pair in combinations(a, 2):
            is_pos = False
            for pos_pair in pos_pair_set:
                if ((pair[0] == pos_pair[0]) and (pair[1] == pos_pair[1])) or (pair[1] == pos_pair[0]) and (
                        pair[0] == pos_pair[1]):
                    is_pos = True
                    break

            if not is_pos:
                neg_pair_set.append(pair)

        self.pos_pair_set = pos_pair_set
        self.neg_pair_set = neg_pair_set

    def forward(self):
        raise NotImplementedError

    def update(self):
        self.optim.zero_grad()
        loss, loss_breakdown = self.forward()
        loss.backward()
        self.optim.step()

        desc = " ".join(['{} {:.3f}'.format(k, v) for k, v in loss_breakdown.items()])
        return desc


class InfoMax(Base):

    def __init__(self, net, optim, batch_size, positive_instance_pairs):
        super().__init__(
            net=net, optim=optim, batch_size=batch_size,
            positive_instance_pairs=positive_instance_pairs
        )

    def forward(self):
        self.divide_pos_neg_pairs()
        if (self.pos_pair_set is None) or \
                ((self.neg_pair_set is None)) or \
                (min(len(self.neg_pair_set), len(self.pos_pair_set)) == 0):
            return {'loss': None}

        num_sample = min(len(self.neg_pair_set), len(self.pos_pair_set))
        pos_pairs = np.asarray(random.sample(self.pos_pair_set, num_sample))
        neg_pairs = np.asarray(random.sample(self.neg_pair_set, num_sample))

        a = torch.as_tensor(self.net.goal_set[pos_pairs[:, 0]], dtype=torch.float32, device=self.net.device)
        b = torch.as_tensor(self.net.goal_set[pos_pairs[:, 1]], dtype=torch.float32, device=self.net.device)
        c = torch.as_tensor(self.net.goal_set[neg_pairs[:, 0]], dtype=torch.float32, device=self.net.device)
        d = torch.as_tensor(self.net.goal_set[neg_pairs[:, 1]], dtype=torch.float32, device=self.net.device)

        abcd = torch.cat([a, b, c, d])
        z_abcd, mu, std = self.net.encode_goal(abcd)
        z_a, z_b, z_c, z_d = torch.chunk(z_abcd, 4)

        loss_infomax, infomax_acc = infomax_loss_fn(z_p0=z_a, z_p1=z_b, z_n0=z_c, z_n1=z_d)

        return {'loss': loss_infomax, 'acc': infomax_acc,
                'mu': mu, 'std': std}


class InfoMin(Base):

    def __init__(self, net, optim, batch_size, positive_instance_pairs):
        super().__init__(
            net=net, optim=optim, batch_size=batch_size,
            positive_instance_pairs=positive_instance_pairs
        )

    def forward(self, mu=None, std=None):

        if mu is None:
            pair = np.asarray(list(self.positive_instance_pairs)).copy()
            np.random.shuffle(pair)
            anc = pair[:self.batch_size, 0]
            pos = pair[:self.batch_size, 1]

            if random.random() < 0.5:
                tmp = anc.copy()
                anc = pos.copy()
                pos = tmp

            anc = torch.as_tensor(self.net.goal_set[anc], dtype=torch.float32,
                                  device=self.net.device)
            pos = torch.as_tensor(self.net.goal_set[pos], dtype=torch.float32,
                                  device=self.net.device)

            z_a, m_a, s_a = self.net.encode_goal(anc)
            with torch.no_grad():
                z_pos, m_p, s_p = self.net.encode_goal(pos)

            mu = torch.cat([m_a, m_p])
            std = torch.cat([s_a, s_p])

        loss = infomin_loss_fn(mu, std)
        return {'loss': loss}



import os, sys
import shutil
import matplotlib.pyplot as plt
import numpy as np
import torch
import torch.nn as nn
import matplotlib


class Reshape(nn.Module):
    def __init__(self, *args):
        super(Reshape, self).__init__()
        self.shape = args

    def forward(self, x):
        return x.view(self.shape)


def get_rgba_from_cmap(cmap, val):
    cmap = matplotlib.cm.get_cmap(cmap)
    return np.asarray(cmap(val)).reshape(1, -1)


def get_device(gpu_id):
    if torch.cuda.is_available():
        device = torch.device(f'cuda:{gpu_id}')
    else:
        device = torch.device(f'cpu')
    return device


def get_coloset_elem_dist(row, arr):
    assert len(arr.shape) == 2
    row = np.asarray(row)
    dist = np.sum((arr - row) ** 2, axis=1, keepdims=True)
    idx_min = dist.argmin(axis=0)[0]
    return dist[idx_min], idx_min


def get_row_index_in_arr(row, arr):
    assert len(arr.shape) == 2
    row = np.asarray(row)
    min_dist, min_dist_idx = get_coloset_elem_dist(row, arr)
    return min_dist_idx if min_dist < 1e-6 else False


def make_path_if_not_exist(path):
    if not os.path.exists(path):
        os.makedirs(path)


def remove_path(path):
    if os.path.exists(path):
        shutil.rmtree(path)


def remove_make_path(path):
    remove_path(path)
    os.makedirs(path)


def copy_move_file(source, target):
    shutil.copy(source, target)


def draw_confusion_matrix(y_pred, y_true):
    # https: // vitalflux.com / accuracy - precision - recall - f1 - score - python - example /
    from sklearn.metrics import confusion_matrix
    from sklearn.metrics import precision_score, recall_score, f1_score, accuracy_score

    conf_matrix = confusion_matrix(y_true=y_true, y_pred=y_pred)
    fig, ax = plt.subplots(figsize=(5, 5))
    ax.matshow(conf_matrix, cmap=plt.cm.Oranges, alpha=0.3)
    for i in range(conf_matrix.shape[0]):
        for j in range(conf_matrix.shape[1]):
            ax.text(x=j, y=i, s=conf_matrix[i, j], va='center', ha='center', size='xx-large')

    plt.xlabel('Predictions', fontsize=18)
    plt.ylabel('Actuals', fontsize=18)
    plt.title('Confusion Matrix', fontsize=18)
    plt.savefig('confusion_matrix.png')
    print('Num. of Obj.: %.3d' % len(y_true))
    print('Num. Pos./ Num. Neg.: {}/{}'.format(y_true.count(1), y_true.count(0)))
    print('Precision: %.3f' % precision_score(y_true, y_pred))
    print('Recall: %.3f' % recall_score(y_true, y_pred))
    print('Accuracy: %.3f' % accuracy_score(y_true, y_pred))
    print('F1 Score: %.3f' % f1_score(y_true, y_pred))


def to_np(x, dtype=np.float32):
    return np.asarray(x, dtype=dtype)


def to_th(x, device, dtype=torch.float32):
    return torch.as_tensor(x, dtype=dtype, device=device)


def get_tianshou_envs(make_env, args):
    import tianshou as ts
    training_env = make_env(mode='training', args=args)
    test_env = make_env(mode='test', args=args)
    train_envs = ts.env.DummyVectorEnv(
        [lambda: make_env(mode='training', args=args) for _ in range(args.training_num)])
    train_envs2 = ts.env.DummyVectorEnv(
        [lambda: make_env(mode='train_for_test', args=args) for _ in range(args.training_num)])
    test_envs = ts.env.DummyVectorEnv(
        [lambda: make_env(mode='test', args=args) for _ in range(args.test_num)])
    return training_env, test_env, train_envs, train_envs2, test_envs

def get_tianshou_logger(log_dir):
    from tianshou.utils import TensorboardLogger
    from torch.utils.tensorboard import SummaryWriter
    return TensorboardLogger(SummaryWriter(log_dir))

def get_encoder_and_decoder(dim, hidden_dim, z_dim, hidden_sizes_for_enc=None):
    import numpy as np
    import torch
    import torch.nn as nn
    from tianshou.utils.net.common import MLP

    if isinstance(dim, tuple) and len(dim) == 3:  # img
        c, h, w = dim
        cnn = nn.Sequential(
            nn.Conv2d(c, 8, kernel_size=4, stride=2), nn.ReLU(inplace=True),
            nn.Conv2d(8, 16, kernel_size=2, stride=1), nn.ReLU(inplace=True),
        )
        with torch.no_grad():
            unflatten_dim = cnn(torch.zeros(1, c, h, w)).shape[1:]
            flatten_dim = int(np.prod(unflatten_dim))

        enc = nn.Sequential(
            cnn,
            nn.Flatten(),
            nn.Linear(flatten_dim, hidden_dim), nn.ReLU(inplace=True),
            # nn.LayerNorm(hidden_dim),
            nn.Linear(hidden_dim, 2 * z_dim)
        )

        dec = nn.Sequential(
            nn.Linear(z_dim, flatten_dim), nn.ReLU(inplace=True),
            Reshape(-1, *unflatten_dim),
            nn.ConvTranspose2d(16, 8, kernel_size=2, stride=1), nn.ReLU(inplace=True),
            nn.ConvTranspose2d(8, c, kernel_size=4, stride=2), nn.ReLU(inplace=True),
        )

    else:
        if len(dim) == 1:
            dim = dim[0]
        if hidden_sizes_for_enc is None:
            hidden_sizes_for_enc = [hidden_dim]*3


        enc = MLP(input_dim=dim, output_dim=2 * z_dim, hidden_sizes=hidden_sizes_for_enc)
        dec = MLP(input_dim=z_dim, output_dim=dim, hidden_sizes=[hidden_dim])

    return enc, dec


def load_pickle(path):
    import pickle
    return pickle.load(open(path, 'rb'))


import copy
import random
import time
import warnings
from typing import Any, Callable, Dict, List, Optional, Union

import gym
import numpy as np
import torch

from tianshou.data import (
    Batch,
    CachedReplayBuffer,
    ReplayBuffer,
    ReplayBufferManager,
    VectorReplayBuffer,
    to_numpy,
    Collector
)
from tianshou.data.batch import _alloc_by_keys_diff
from tianshou.env import BaseVectorEnv, DummyVectorEnv
from tianshou.policy import BasePolicy

T = 100


def get_new_trajactory_len(buffers, env_id, ep_len, relabelling_strategy, goal_info_dict):
    env_buffer = buffers[env_id]
    env_buffer_len = env_buffer.last_index[0] + 1
    traj_len = ep_len[env_id]
    obs_index_range = np.arange(env_buffer_len - traj_len, env_buffer_len) % len(env_buffer)
    original_trajectory = env_buffer[obs_index_range]

    achieved_goal_id = original_trajectory.obs_next.achieved_goal_id
    desired_goal_id = original_trajectory.obs_next.desired_goal_id
    # print(desired_goal_id)
    assert len(desired_goal_id) > 0
    assert len(np.unique(desired_goal_id)) == 1
    unique_desired_goal_id = np.unique(desired_goal_id)[0]
    achieved_goal_id = np.unique(achieved_goal_id[achieved_goal_id >= 0])

    max_rew = original_trajectory.obs_next.max_rew[0]

    if len(achieved_goal_id) == 0:
        return None, None
    # if relabelling_strategy == 'advantage':
    #     s0_id = original_trajectory.obs.state_id[0]
    #     achieved_goal_id_his = original_trajectory.obs.achieved_goal_id_his[0]
    #     s0_id = [s0_id] * len(achieved_goal_id)
    #     achieved_goal_id_his = [achieved_goal_id_his] * len(achieved_goal_id)
    #     obs = Batch({'state_id': s0_id, 'desired_goal_id': achieved_goal_id, 'achieved_goal_id_his': achieved_goal_id_his})
    #     with torch.no_grad():
    #         q, _ = q_net(obs, state=None, info=None)
    #         q = torch.amax(q, dim=-1).cpu().numpy()

    new_trajactory_len = []
    for index, achieved_goal_id_i in enumerate(achieved_goal_id):

        length_candidate = np.where(original_trajectory.obs_next.achieved_goal_id == achieved_goal_id_i)[0]
        if relabelling_strategy == 'min_length':
            length = np.amin(length_candidate)
        elif relabelling_strategy == 'cher':
            if goal_info_dict[unique_desired_goal_id]['room'] != goal_info_dict[achieved_goal_id_i]['room']:
                continue
            length = np.amin(length_candidate)
        elif relabelling_strategy == 'random':
            length = random.choice(length_candidate)
        # elif relabelling_strategy == 'advantage':
        #     length = np.amin(length_candidate)
        #     q_i = q[index]
        #     cumulative_reward = -(length - 1) * (1 / T) + max_rew
        #     if q_i > cumulative_reward:
        #         continue
        else:
            raise NotImplementedError
        new_trajactory_len.append(length)

    if len(new_trajactory_len) == 0:
        return None, None

    return new_trajactory_len, original_trajectory


def get_collector(rl):
    assert rl in ['DQN', 'HER', 'CHER', 'GHRL']
    return {'DQN': Collector, 'HER': HERCollector, 'CHER': CHERCollector, 'GHRL': GHRLCollector}[rl]


class HERCollector(Collector):
    def __init__(self,
                 policy: BasePolicy,
                 env: Union[gym.Env, BaseVectorEnv],
                 buffer: Optional[ReplayBuffer] = None,
                 preprocess_fn: Optional[Callable[..., Batch]] = None,
                 exploration_noise: bool = False,
                 max_len: int = 1e6,
                 ):
        super(HERCollector, self).__init__(
            policy,
            env,
            buffer,
            preprocess_fn,
            exploration_noise
        )
        self.max_len = max_len
        self.relabelling_strategy = 'random'
        self.goal_info_dict = env.workers[0].env.goal_info_dict

    def collect(
            self,
            n_step: Optional[int] = None,
            n_episode: Optional[int] = None,
            random: bool = False,
            render: Optional[float] = None,
            no_grad: bool = True,
    ) -> Dict[str, Any]:
        """Collect a specified number of step or episode.

        To ensure unbiased sampling result with n_episode option, this function will
        first collect ``n_episode - env_num`` episodes, then for the last ``env_num``
        episodes, they will be collected evenly from each env.

        :param int n_step: how many steps you want to collect.
        :param int n_episode: how many episodes you want to collect.
        :param bool random: whether to use random policy for collecting data. Default
            to False.
        :param float render: the sleep time between rendering consecutive frames.
            Default to None (no rendering).
        :param bool no_grad: whether to retain gradient in policy.forward(). Default to
            True (no gradient retaining).

        .. note::

            One and only one collection number specification is permitted, either
            ``n_step`` or ``n_episode``.

        :return: A dict including the following keys

            * ``n/ep`` collected number of episodes.
            * ``n/st`` collected number of steps.
            * ``rews`` array of episode reward over collected episodes.
            * ``lens`` array of episode length over collected episodes.
            * ``idxs`` array of episode start index in buffer over collected episodes.
            * ``rew`` mean of episodic rewards.
            * ``len`` mean of episodic lengths.
            * ``rew_std`` standard error of episodic rewards.
            * ``len_std`` standard error of episodic lengths.
        """
        assert not self.env.is_async, "Please use AsyncCollector if using async venv."
        if n_step is not None:
            assert n_episode is None, (
                f"Only one of n_step or n_episode is allowed in Collector."
                f"collect, got n_step={n_step}, n_episode={n_episode}."
            )
            assert n_step > 0
            if not n_step % self.env_num == 0:
                warnings.warn(
                    f"n_step={n_step} is not a multiple of #env ({self.env_num}), "
                    "which may cause extra transitions collected into the buffer."
                )
            ready_env_ids = np.arange(self.env_num)
        elif n_episode is not None:
            assert n_episode > 0
            ready_env_ids = np.arange(min(self.env_num, n_episode))
            self.data = self.data[:min(self.env_num, n_episode)]
        else:
            raise TypeError(
                "Please specify at least one (either n_step or n_episode) "
                "in AsyncCollector.collect()."
            )

        start_time = time.time()

        step_count = 0
        episode_count = 0
        episode_rews = []
        episode_lens = []
        episode_start_indices = []

        while True:
            assert len(self.data) == len(ready_env_ids)
            # restore the state: if the last state is None, it won't store
            last_state = self.data.policy.pop("hidden_state", None)

            # get the next action
            if random:
                self.data.update(
                    act=[self._action_space[i].sample() for i in ready_env_ids]
                )
            else:
                if no_grad:
                    with torch.no_grad():  # faster than retain_grad version
                        # self.data.obs will be used by agent to get result
                        result = self.policy(self.data, last_state)
                else:
                    result = self.policy(self.data, last_state)
                # update state / act / policy into self.data
                policy = result.get("policy", Batch())
                assert isinstance(policy, Batch)
                state = result.get("state", None)
                if state is not None:
                    policy.hidden_state = state  # save state into buffer
                act = to_numpy(result.act)
                if self.exploration_noise:
                    act = self.policy.exploration_noise(act, self.data)
                self.data.update(policy=policy, act=act)

            # get bounded and remapped actions first (not saved into buffer)
            action_remap = self.policy.map_action(self.data.act)
            # step in env
            result = self.env.step(action_remap, ready_env_ids)  # type: ignore
            obs_next, rew, done, info = result

            self.data.update(obs_next=obs_next, rew=rew, done=done, info=info)
            if self.preprocess_fn:
                self.data.update(
                    self.preprocess_fn(
                        obs_next=self.data.obs_next,
                        rew=self.data.rew,
                        done=self.data.done,
                        info=self.data.info,
                        policy=self.data.policy,
                        env_id=ready_env_ids,
                    )
                )

            if render:
                self.env.render()
                if render > 0 and not np.isclose(render, 0):
                    time.sleep(render)

            # add data into the buffer
            ptr, ep_rew, ep_len, ep_idx = self.buffer.add(
                self.data, buffer_ids=ready_env_ids
            )

            # collect statistics
            step_count += len(ready_env_ids)

            if np.any(done):
                env_ind_local = np.where(done)[0]
                env_ind_global = ready_env_ids[env_ind_local]
                episode_count += len(env_ind_local)
                episode_lens.append(ep_len[env_ind_local])
                episode_rews.append(ep_rew[env_ind_local])
                episode_start_indices.append(ep_idx[env_ind_local])
                # now we copy obs_next to obs, but since there might be
                # finished episodes, we have to reset finished envs first.
                obs_reset = self.env.reset(env_ind_global)
                if self.preprocess_fn:
                    obs_reset = self.preprocess_fn(
                        obs=obs_reset, env_id=env_ind_global
                    ).get("obs", obs_reset)
                self.data.obs_next[env_ind_local] = obs_reset
                for i in env_ind_local:
                    self._reset_state(i)

                # remove surplus env id from ready_env_ids
                # to avoid bias in selecting environments
                if n_episode:
                    surplus_env_num = len(ready_env_ids) - (n_episode - episode_count)
                    if surplus_env_num > 0:
                        mask = np.ones_like(ready_env_ids, dtype=bool)
                        mask[env_ind_local[:surplus_env_num]] = False
                        ready_env_ids = ready_env_ids[mask]
                        self.data = self.data[mask]

                # use HER to create more trajectory
                for env_id in env_ind_global:  # enumerate env
                    # get recently collected data from buffer

                    new_trajactory_len, original_trajectory = get_new_trajactory_len(
                        buffers=self.buffer.buffers, env_id=env_id, ep_len=ep_len,
                        relabelling_strategy=self.relabelling_strategy,
                        goal_info_dict=self.goal_info_dict,
                    )

                    if new_trajactory_len is None:
                        continue

                    for length in new_trajactory_len:  # use different mid goal to resample
                        if length > self.max_len: continue
                        trajectory = Batch(original_trajectory[:length + 1], copy=True)

                        new_goal = trajectory.obs_next.achieved_goal_id[length]

                        new_goals = np.repeat([new_goal], length + 1, axis=0)
                        trajectory.obs.desired_goal_id = new_goals
                        trajectory.obs_next.desired_goal_id = new_goals
                        trajectory.rew[-1] = trajectory.obs_next.max_rew[0]
                        trajectory.done[-1] = True
                        for transition in trajectory:
                            self.buffer.buffers[env_id].add(transition)

                # ---------------------------------------------
            self.data.obs = self.data.obs_next

            if (n_step and step_count >= n_step) or \
                    (n_episode and episode_count >= n_episode):
                break

        # generate statistics
        self.collect_step += step_count
        self.collect_episode += episode_count
        self.collect_time += max(time.time() - start_time, 1e-9)

        if n_episode:
            self.data = Batch(
                obs={}, act={}, rew={}, done={}, obs_next={}, info={}, policy={}
            )
            self.reset_env()

        if episode_count > 0:
            rews, lens, idxs = list(
                map(
                    np.concatenate,
                    [episode_rews, episode_lens, episode_start_indices]
                )
            )
            rew_mean, rew_std = rews.mean(), rews.std()
            len_mean, len_std = lens.mean(), lens.std()
        else:
            rews, lens, idxs = np.array([]), np.array([], int), np.array([], int)
            rew_mean = rew_std = len_mean = len_std = 0

        return {
            "n/ep": episode_count,
            "n/st": step_count,
            "rews": rews,
            "lens": lens,
            "idxs": idxs,
            "rew": rew_mean,
            "len": len_mean,
            "rew_std": rew_std,
            "len_std": len_std,
        }


class CHERCollector(HERCollector):
    def __init__(self,
                 policy: BasePolicy,
                 env: Union[gym.Env, BaseVectorEnv],
                 buffer: Optional[ReplayBuffer] = None,
                 preprocess_fn: Optional[Callable[..., Batch]] = None,
                 exploration_noise: bool = False,
                 max_len: int = 1e6,
                 ):
        super(CHERCollector, self).__init__(
            policy,
            env,
            buffer,
            preprocess_fn,
            exploration_noise
        )
        self.max_len = max_len
        self.relabelling_strategy = 'cher'


class GHRLCollector(HERCollector):
    def __init__(self,
                 policy: BasePolicy,
                 env: Union[gym.Env, BaseVectorEnv],
                 buffer: Optional[ReplayBuffer] = None,
                 preprocess_fn: Optional[Callable[..., Batch]] = None,
                 exploration_noise: bool = False,
                 max_len: int = 1e6,
                 ):
        super(GHRLCollector, self).__init__(
            policy,
            env,
            buffer,
            preprocess_fn,
            exploration_noise
        )
        self.max_len = max_len
        self.relabelling_strategy = 'min_length'
